/*************************************************************************
***	Authentication, authorization, accounting + firewalling package
***	(c) 1998-2002 Anton Vinokurov <anton@netams.com>
***	(c) 2002-2005 NeTAMS Development Team
***	All rights reserved. See 'Copying' file included in distribution
***	For latest version and more info, visit this project web page
***	located at http://www.netams.com
***
*************************************************************************/
/* $Id: ds_netflow.c,v 1.68.2.4 2005/11/24 15:45:38 jura Exp $ */

#include "netams.h"
#include "netflow.h"

//////////////////////////////////////////////////////////////////////////
unsigned int	sDS_AcctNetFlowV5(ServiceDS_cfg* cfg, NFSource *nfSource, void *packet, time_t time);
unsigned int	sDS_AcctNetFlowV9(ServiceDS_cfg* cfg, NFSource *nfSource, u_char *packet, time_t time,unsigned int len);
void		ds_netflow_cancel(void *ptr);
void		processV9Template(u_char *packet, NFSource *nfSource, unsigned int len);
void		processV9Data(u_char *packet, NFSource *nfSource, Flow *flow);
u_int64_t	v9ConvertIntField(void *field, int size, int tohbo);
void 		createFlowAttribute(Flow *flow, const unsigned short attribute, void **value);

//////////////////////////////////////////////////////////////////////////
void	*ds_netflow(void *ss) {
	Service 		*s;
        ServiceDS_cfg		*cfg;
        struct sockaddr_in 	ds;
        int 			socketid, reuseaddr, status;
	int	 		len;
        void	 		*packet;
        struct timeval 		start;
        NFSource 		*nfSource;
	socklen_t		size_ds;

	s		= (Service *)ss;
	cfg		= (ServiceDS_cfg *)s->cfg;
	packet		= cfg->packet;
	size_ds		= sizeof(ds);
        reuseaddr	= 1;

	socketid	= socket(PF_INET, SOCK_DGRAM, 0);
	
        if( socketid < 0 ) { 
		aLog(D_ERR, "failed to create netflow listening socket: %s", strerror(errno));
		return NULL;
	}
	
        bzero(&ds, size_ds);
	ds.sin_family		= AF_INET;
        ds.sin_addr.s_addr	= cfg->l_addr.s_addr;
        ds.sin_port		= htons(cfg->port);

        // set options
	status	= bigsockbuf(socketid, SO_RCVBUF, FT_SO_RCV_BUFSIZE);
	if( status < 0 ) {
		aLog(D_ERR, "failed to setsockopt receive buffer: %s", strerror(errno));
		return NULL;
	} else {
		aLog(D_INFO, "receive bufer set to %u\n", status);
	}
	
        status = setsockopt(socketid, SOL_SOCKET, SO_REUSEADDR, (void *)&reuseaddr, sizeof(reuseaddr));
	if( status < 0 ) {
                aLog(D_ERR, "failed to setsockopt netflow listening socket: %s", strerror(errno));
                return NULL;
        }

        // bind socket with serv_addr
        status = bind(socketid, (struct sockaddr *)&ds, size_ds);
        if( status < 0 ) { 
		aLog(D_ERR, "failed to bind netflow listening socket: %s", strerror(errno)); 
		return NULL; 
	}
	SET_POLL( socketid );

	pthread_cleanup_push(ds_netflow_cancel, (void *)&socketid);
	pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, NULL);
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
	
	cfg->ds_flow = new Flow(s->instance,cfg->max_flow_slots);

	char buffer[32];
	aLog(D_INFO,"NetFlow packet processing for data-source:%u initialized\n", s->instance);
	aLog(D_INFO,"Listening on %s:%u\n", inet_ntop(AF_INET, &(cfg->l_addr), buffer, 32), cfg->port);
	
        while( 1 ) {
		CHECK_POLL( status );
		if( !status ) continue;

		if((pfd.revents & POLLIN) == 0)
		    continue;

		// starting from here we need additional sanity checks
		len = recvfrom(socketid, packet, MAX_PKT_SIZE, 0, (struct sockaddr *)&ds, &size_ds);

		if(len == -1) {
		    aLog(D_ERR, "Can not read from network\n");
		    continue;
		}
		
		if(len < 2) {
		    aLog(D_ERR, "Junk packet received, dropping.\n");
		    continue;
		}
		
		//look for source
		for(nfSource = cfg->nfroot; nfSource != NULL; nfSource = nfSource->next)
			if(ds.sin_addr.s_addr == nfSource->src_addr)  break;

		if( !nfSource ) {
			cfg->total_errors++;
			aLog(D_WARN, "NetFlow packet from unknown source: %s\n", inet_ntop(AF_INET, &(ds.sin_addr), buffer, 32)); 
			continue; 
		}
		
                netams_gettimeofday(&start, NULL);
		switch(ntohs(*((u_int16_t *)packet))) {
		    case FLOW_VERSION_5:
            		sDS_AcctNetFlowV5(cfg, nfSource, packet, (time_t)start.tv_sec);
			break;
		    case FLOW_VERSION_9:
            		sDS_AcctNetFlowV9(cfg, nfSource, (u_char *)packet, (time_t)start.tv_sec, len);
			break;
		    default:
			aLog(D_WARN, " NetFlow v%d packet unsupported\n", ntohs(*((u_int16_t *)packet)));
			cfg->total_errors++;
			nfSource->errors++;
			break;
		}
		
		nfSource->packets++;
		nfSource->timestamp = cfg->timestamp = start.tv_sec;
                
		sDSMeasure(cfg, &start, len);
    	}

	pthread_cleanup_pop(0);
	return NULL;
}
//////////////////////////////////////////////////////////////////////////
void ds_netflow_cancel(void *ptr) {
	int socketid = *((int *)ptr);
	
	shutdown(socketid, SHUT_RDWR);
        close( socketid );
}

unsigned int sDS_AcctNetFlowV9(ServiceDS_cfg *cfg, NFSource *nfSource, u_char *packet, time_t time,unsigned int len) {
	struct	NetFlowV9Header		*nfHeader;
	struct	NetFlowV9FlowSetHeader	*nfFlowSetHeader;
	unsigned int			offset;

	if(len < sizeof(struct NetFlowV9Header)) {
	    aLog(D_ERR, "Junk packet received, dropping.\n");
	    return 1;
	}
	
	nfHeader			= (struct NetFlowV9Header *)packet;
	// Count is strange. It looks like padding to 32. Will open TAC
	// to clarify that 'cause docs do not cover that
	nfHeader->count			= ntohs(nfHeader->count);
	nfHeader->sysUpTime		= ntohl(nfHeader->sysUpTime);
	nfHeader->unixSeconds		= ntohl(nfHeader->unixSeconds);
	nfHeader->packageSequence	= ntohl(nfHeader->packageSequence);
	nfHeader->sourceId		= ntohl(nfHeader->sourceId);

	if( nfSource->seq_id != 0 && nfSource->seq_id != (nfHeader->packageSequence - 1) ) { 
	    aLog(D_WARN, "NF:%u awaited seq %lu, received %lu from %02X (%u flows are lost)\n", 
		cfg->id, nfSource->seq_id+1, nfHeader->packageSequence, 
		nfHeader->sourceId, nfHeader->packageSequence - nfSource->seq_id - 1);
		nfSource->errors++;
		cfg->total_errors++;
	}
	
	nfSource->seq_id = nfHeader->packageSequence;  // storing sequence

	packet	+= sizeof(struct NetFlowV9Header);
	len	-= sizeof(struct NetFlowV9Header);
	
	offset = 0;
	while(offset < len) {
	    if(offset + sizeof(struct NetFlowV9FlowSetHeader) > len) {
		aLog(D_ERR, "Unexpected end of packet, stop processing.\n");
		return 1;
	    }
	    nfFlowSetHeader		= (struct NetFlowV9FlowSetHeader *)(packet + offset);
	    nfFlowSetHeader->flowSetId	= ntohs(nfFlowSetHeader->flowSetId);
	    nfFlowSetHeader->length	= ntohs(nfFlowSetHeader->length);

	    if(offset + nfFlowSetHeader->length > len) {
		aLog(D_ERR, "Unexpected end of packet, stop processing.\n");
		return 1;
	    }
	    if(nfFlowSetHeader->flowSetId == FLOW_VERSION_9_TMPL_ID) {
		processV9Template(packet + offset + sizeof(struct NetFlowV9FlowSetHeader), nfSource, nfFlowSetHeader->length);
	    } else if(nfFlowSetHeader->flowSetId == FLOW_VERSION_9_TMPL_OPT_ID) {
		aLog(D_INFO, "Received V9 options template. Not yet implemented\n");
	    } else if(nfFlowSetHeader->flowSetId > 255) {
		if(nfFlowSetHeader->flowSetId == nfSource->nfTemplate.header.templateId) {
		    Flow *flow = cfg->ds_flow; 
		    aDebug(DEBUG_FLOW, "Received v9 data packets\n");
		    processV9Data(packet + offset + sizeof(struct NetFlowV9FlowSetHeader), nfSource, flow);
		    cfg->FE->DoSend(flow);
    		    flow->reuse();
		} else
		    aLog(D_INFO, "Received data for unknown template, ignoring\n");
	    } else {
		aLog(D_WARN, "Received unknown template, ignoring.\n");
	    }
	    offset += nfFlowSetHeader->length;
	}
	
	return 0;
}

void processV9Template(u_char *packet, NFSource *nfSource, unsigned int len) {
    struct NetFlowV9TemplateHeader	*nfTemplateHeader;
    unsigned int			i;
    
    // I know, this is weird, but by the first we will support only one template
    nfTemplateHeader = (struct NetFlowV9TemplateHeader *)packet;
    nfTemplateHeader->templateId = ntohs(nfTemplateHeader->templateId);
    nfTemplateHeader->fieldCount = ntohs(nfTemplateHeader->fieldCount);
    if(nfSource->nfTemplate.fields == NULL || nfSource->nfTemplate.header.templateId != nfTemplateHeader->templateId) {
	// new template received
	if(nfSource->nfTemplate.fields != NULL)
	    aFree(nfSource->nfTemplate.fields);
	if((nfSource->nfTemplate.fields = (struct NetFlowV9FieldInfo *)aMalloc(sizeof(struct NetFlowV9FieldInfo)*nfTemplateHeader->fieldCount)) == NULL) {
	    aLog(D_ERR, "Can not allocate memory for template.\n");
	    return;
	}
	bcopy(packet + sizeof(struct NetFlowV9TemplateHeader), nfSource->nfTemplate.fields, sizeof(struct NetFlowV9FieldInfo)*nfTemplateHeader->fieldCount);
	for(i = 0; i < nfTemplateHeader->fieldCount; i++) {
	    nfSource->nfTemplate.fields[i].type		= ntohs(nfSource->nfTemplate.fields[i].type);
	    nfSource->nfTemplate.fields[i].length	= ntohs(nfSource->nfTemplate.fields[i].length);
	}
	nfSource->nfTemplate.header = *nfTemplateHeader;
    }
}

void createFlowAttribute(Flow *flow, const unsigned short attribute, void **value) {
    if(*value == NULL)
	*value = flow->put(attribute);
    return;
}

void processV9Data(u_char *packet, NFSource *nfSource, Flow *flow) {
    unsigned int	i, offset;

    struct flow_info_value		*flow_info	= NULL;
    struct ipv4_info_value		*ipv4_info	= NULL;
    struct tcp_info_value		*tcp_info	= NULL;
    struct as_info_value		*as_info	= NULL;
    struct ifindex_info_value		*ifindex_info	= NULL;
    struct multicast_info_value		*mcast_info	= NULL;
    struct nexthop_info_value		*nexthop_info	= NULL;

    flow_info = (struct flow_info_value *)flow->put(ATTR_FLOW_INFO);

    // iterating over fields
    for(i = 0, offset = 0; i < nfSource->nfTemplate.header.fieldCount; i++) {
	switch( nfSource->nfTemplate.fields[i].type ) {
	    case NFV9_IN_BYTES:
		flow_info->octets = v9ConvertIntField(packet + offset, nfSource->nfTemplate.fields[i].length, 1);
		break;
	    case NFV9_IN_PKTS:
		flow_info->packets = v9ConvertIntField(packet + offset, nfSource->nfTemplate.fields[i].length, 1);
		break;
	    case NFV9_DIRECTION:
		flow_info->direction = (u_char)v9ConvertIntField(packet + offset, nfSource->nfTemplate.fields[i].length, 0);
		break;
	    case NFV9_PROTOCOL:
		createFlowAttribute(flow, ATTR_IPV4_INFO, (void **)&ipv4_info);
		ipv4_info->ip_p = v9ConvertIntField(packet + offset, nfSource->nfTemplate.fields[i].length, 0);
		break;
	    case NFV9_SRC_TOS:
		createFlowAttribute(flow, ATTR_IPV4_INFO, (void **)&ipv4_info);
		ipv4_info->ip_tos = v9ConvertIntField(packet + offset, nfSource->nfTemplate.fields[i].length, 0);
		break;
	    case NFV9_IPV4_SRC_ADDR:
		createFlowAttribute(flow, ATTR_IPV4_INFO, (void **)&ipv4_info);
		ipv4_info->ip_src.s_addr = v9ConvertIntField(packet + offset, nfSource->nfTemplate.fields[i].length, 0);
		break;
	    case NFV9_IPV4_DST_ADDR:
		createFlowAttribute(flow, ATTR_IPV4_INFO, (void **)&ipv4_info);
		ipv4_info->ip_dst.s_addr = v9ConvertIntField(packet + offset, nfSource->nfTemplate.fields[i].length, 0);
		break;
	    case NFV9_SRC_AS:
		createFlowAttribute(flow, ATTR_AS_INFO, (void **)&as_info);
		as_info->as_src = v9ConvertIntField(packet + offset, nfSource->nfTemplate.fields[i].length, 0);
		break;
	    case NFV9_DST_AS:
		createFlowAttribute(flow, ATTR_AS_INFO, (void **)&as_info);
		as_info->as_dst = v9ConvertIntField(packet + offset, nfSource->nfTemplate.fields[i].length, 0);
		break;
	    case NFV9_INPUT_SNMP:
		createFlowAttribute(flow, ATTR_IFINDEX_INFO, (void **)&ifindex_info);
		ifindex_info->if_in = v9ConvertIntField(packet + offset, nfSource->nfTemplate.fields[i].length, 0);
		break;
	    case NFV9_OUTPUT_SNMP:
		createFlowAttribute(flow, ATTR_IFINDEX_INFO, (void **)&ifindex_info);
		ifindex_info->if_out = v9ConvertIntField(packet + offset, nfSource->nfTemplate.fields[i].length, 0);
		break;
	    case NFV9_L4_SRC_PORT:
		createFlowAttribute(flow, ATTR_TCP_INFO, (void **)&tcp_info);
		tcp_info->src_port = v9ConvertIntField(packet + offset, nfSource->nfTemplate.fields[i].length, 0);
		break;
	    case NFV9_L4_DST_PORT:
		createFlowAttribute(flow, ATTR_TCP_INFO, (void **)&tcp_info);
		tcp_info->dst_port = v9ConvertIntField(packet + offset, nfSource->nfTemplate.fields[i].length, 0);
		break;
	    case NFV9_MUL_DST_PKTS:
		createFlowAttribute(flow, ATTR_MULTICAST_INFO, (void **)&mcast_info);
		mcast_info->dst_packets = v9ConvertIntField(packet + offset, nfSource->nfTemplate.fields[i].length, 1);
		break;
	    case NFV9_MUL_DST_BYTES:
		createFlowAttribute(flow, ATTR_MULTICAST_INFO, (void **)&mcast_info);
		mcast_info->dst_bytes = v9ConvertIntField(packet + offset, nfSource->nfTemplate.fields[i].length, 1);
		break;
	    case NFV9_MUL_IGMP_TYPE:
		createFlowAttribute(flow, ATTR_MULTICAST_INFO, (void **)&mcast_info);
		mcast_info->igmp_type = (u_char)v9ConvertIntField(packet + offset, nfSource->nfTemplate.fields[i].length, 0);
		break;
	    case NFV9_IPV4_NEXT_HOP:
		createFlowAttribute(flow, ATTR_NEXTHOP_INFO, (void **)&nexthop_info);
		nexthop_info->ipv4_nexthop.s_addr = v9ConvertIntField(packet + offset, nfSource->nfTemplate.fields[i].length, 0);
		break;
	    case NFV9_BGP_IPV4_NEXT_HOP:
		createFlowAttribute(flow, ATTR_NEXTHOP_INFO, (void **)&nexthop_info);
		nexthop_info->bgpv4_nexthop.s_addr = v9ConvertIntField(packet + offset, nfSource->nfTemplate.fields[i].length, 0);
		break;
	    default:
		break;
	}
	offset += nfSource->nfTemplate.fields[i].length;
    }
}

u_int64_t v9ConvertIntField(void *field, int size, int tohbo) {
    switch( size ) {
	case 1:
	    return *((u_char *)field);
	case 2:
	    return tohbo ? ntohs(*((u_int16_t *)field)) : *((u_int16_t *)field);
	case 4:
	    return tohbo ? ntohl(*((u_int32_t *)field)) : *((u_int32_t *)field);
	case 8:
	    return tohbo ? ntohll(*((u_int64_t *)field)) : *((u_int64_t *)field);
	default:
	    return 0;
    }
}

unsigned int sDS_AcctNetFlowV5(ServiceDS_cfg* cfg, NFSource *nfSource, void *packet, time_t time) {
	IPStat5Msg			*msg;
	IPFlow5Stat			*v5flow;
	Flow5StatHdr			*hdr;
	
	struct flow_info_value		*flow_info;
	struct ipv4_info_value		*ipv4_info;
	struct tcp_info_value		*tcp_info;
	struct as_info_value		*as_info;
	struct ifindex_info_value	*ifindex_info;
	
	FlowEngine	*FE		= cfg->FE;
	Flow		*flow		= cfg->ds_flow;

	msg	= (IPStat5Msg *)packet;
	hdr	= &msg->header;
	
	hdr->count         = ntohs(hdr->count);
	hdr->SysUptime     = ntohl(hdr->SysUptime);
	hdr->unix_secs     = ntohl(hdr->unix_secs);
	hdr->unix_nsecs    = ntohl(hdr->unix_nsecs);
	hdr->flow_sequence = ntohl(hdr->flow_sequence);

	aDebug(DEBUG_FLOW, " cfg seq %lu, fl seq %lu, %d records:\n", nfSource->seq_id, hdr->flow_sequence, hdr->count);

	if( !nfSource->seq_id )
		nfSource->seq_id = hdr->flow_sequence;
	else if( nfSource->seq_id != hdr->flow_sequence ) { 
		aLog(D_WARN, "NF:%u awaited seq %lu, received %lu from %02X (%u flows are lost)\n",
			cfg->id, nfSource->seq_id, hdr->flow_sequence,
			hdr->engine_id, hdr->flow_sequence-nfSource->seq_id);
			
		if( hdr->count > V5FLOWS_PER_PAK ) 
			aLog(D_WARN, "NF:%u packet with wrong number of flows (>V5FLOWS_PER_PAK), received from %02X \n", cfg->id, hdr->engine_id);
		nfSource->seq_id = hdr->flow_sequence; // restoring back
		nfSource->errors++;
		cfg->total_errors++;
	}
		
	//update counters
	nfSource->seq_id	+= hdr->count;
	nfSource->flows		+= hdr->count;
	cfg->total_flows	+= hdr->count;

	aDebug(DEBUG_FLOW, " NetFlow v5:%02X packet ds:%u seq %lu/%lu with %u records:\n",
		hdr->engine_id, cfg->id, hdr->flow_sequence, nfSource->seq_id, hdr->count);

	for (int i = 0; i < hdr->count; i++) {
		v5flow = &msg->records[i];

		flow->reuse();

		flow_info		= (struct flow_info_value*)flow->put(ATTR_FLOW_INFO);
		flow_info->packets	= ntohl(v5flow->dPkts);
		flow_info->octets	= ntohl(v5flow->dOctets);

		if(cfg->flags == DS_CLOCK_REMOTE) { //e.g == CLOCK_REMOTE for now 
			//use flow time
			flow_info->flow_first	= hdr->unix_secs + ntohl(v5flow->First)/1000 - hdr->SysUptime/1000;
			flow_info->flow_last	= hdr->unix_secs + ntohl(v5flow->Last)/1000 - hdr->SysUptime/1000;
		} else {
			flow_info->flow_first	= time - (ntohl(v5flow->Last) - ntohl(v5flow->First))/1000;
			flow_info->flow_last	= time;
		}
			
		ipv4_info		= (struct ipv4_info_value*)flow->put(ATTR_IPV4_INFO);
		ipv4_info->ip_p		= v5flow->prot;
		ipv4_info->ip_tos	= v5flow->tos;
		ipv4_info->ip_src	= v5flow->srcaddr;
		ipv4_info->ip_dst	= v5flow->dstaddr;

		if(ipv4_info->ip_p == IPPROTO_TCP || ipv4_info->ip_p == IPPROTO_UDP) {
			tcp_info		= (struct tcp_info_value*)flow->put(ATTR_TCP_INFO);
			tcp_info->src_port	= v5flow->srcport;
			tcp_info->dst_port	= v5flow->dstport;
		}
			
		as_info		= (struct as_info_value*)flow->put(ATTR_AS_INFO);
		as_info->as_src	= v5flow->src_as;
		as_info->as_dst	= v5flow->dst_as;
			
		ifindex_info	= (struct ifindex_info_value*)flow->put(ATTR_IFINDEX_INFO);
		ifindex_info->if_in	= v5flow->input;
		ifindex_info->if_out	= v5flow->output;

		FE->DoSend(flow);		
	}

	return 0;
}

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

