/*************************************************************************
***	Authentication, authorization, accounting + firewalling package
***	(c) 1998-2002 Anton Vinokurov <anton@netams.com>
***	(c) 2002-2005 NeTAMS Development Team
***	All rights reserved. See 'Copying' file included in distribution
***	For latest version and more info, visit this project web page
***	located at http://www.netams.com
***
*************************************************************************/
/* $Id: mutex.c,v 1.5 2005/08/30 07:11:42 jura Exp $ */

#include "netams.h"

/////////////////////////////////////////////////////////////////////////////////////////
#if defined(DEBUG) && defined(MUTEX_DEBUG)

#define MUTEX_HASH_SIZE   0x100
#define MUTEX_HASH_MASK   (MUTEX_HASH_SIZE-1)
#define MUTEX_HASH(ptr)   ((unsigned)ptr>>4)&MUTEX_HASH_MASK

#define MAX_THREADS	10

const char* lock_name[2] = { "MUTEX", "RWLOCK" };
const char* lock_type[3] = { "lock", "rwlock", "rdlock" };

#define getLockName(type) lock_name[(type&0x0F)-1]
#define getLockType(type) lock_type[((type&0xF0)>>4)-1]

typedef struct mutex_unit {
	void *ptr;
	u_char type;
	pthread_t created;
	pthread_t used[MAX_THREADS];
	mutex_unit *next;
} mutex_unit;

mutex_unit **mutex_hash=NULL;  //this is hash were we store information about operated mutexes

mutex_unit *getMutexUnit(void *ptr, u_char flags); //ADD, REMOVE
pthread_t getLock(mutex_unit *m);
#else
void aMutexDebugInit() {};
void aMutexDebugRelease() {};
#endif


#if defined(DEBUG) && defined(MUTEX_DEBUG)
pthread_mutex_t mutex_lock = PTHREAD_MUTEX_INITIALIZER;

void aMutexDebugInit() {
	if(!mutex_hash) mutex_hash=(mutex_unit**)calloc(MUTEX_HASH_SIZE,sizeof(mutex_unit*));
	aLog(D_INFO, "Mutexes Debugging Initialized\n");
}

void aMutexDebugRelease(){
	mutex_unit *tmp, *ptr;
	
	aLog(D_INFO, "Mutexes Debugging Stopped\n");

	///clear hash
	for(unsigned i=0;i<MUTEX_HASH_SIZE;i++) {
		for(ptr=mutex_hash[i];ptr!=NULL;) {
			tmp=ptr;
			ptr=ptr->next;
			free(tmp);
		}
	}
	free(mutex_hash);
}

mutex_unit *getMutexUnit(void *ptr, u_char flags) {
	mutex_unit *m, *p=NULL;

	if(!mutex_hash) 
		mutex_hash=(mutex_unit**)calloc(MUTEX_HASH_SIZE,sizeof(mutex_unit*));
	
	unsigned idx = MUTEX_HASH(ptr);

	for(m=mutex_hash[idx];m!=NULL;m=m->next) {
		if(m->ptr == ptr) break;
		p=m;
	}
	
	if(flags == ADD) { //add 2 hash
		if(m) return m;
		m=(mutex_unit*)calloc(1,sizeof(mutex_unit));
		m->ptr=ptr;

		m->next = mutex_hash[idx];
		mutex_hash[idx] = m;
	} else { //remove from hash
		if(!m) return NULL;
		if(mutex_hash[idx] == m) mutex_hash[idx] = m->next;
		else p->next = m->next;
		free(m);
		m = NULL;
	}

	return m;
}

pthread_t getLock(mutex_unit *m, u_char flags) {
	pthread_t caller=pthread_self();
	pthread_t where=0;
	u_char i;
	
	for(i=0; i< MAX_THREADS; i++) {
		if(m->used[i] == caller) {
			where = caller;
			break;
		}
	}

	if(flags==ADD) {
		if(!where) {
			for(i=0; i< MAX_THREADS; i++){
				if(m->used[i]==0) {
					m->used[i]=caller;
					return m->used[i];
				}
			}
			aLog(D_WARN, "MUTEX Debug: Not enough slots: Increase MAX_THREADS(%u) and recompile.\n", MAX_THREADS);
		}
	} else { //REMOVE
		if(where) m->used[i]=0;
	}
	
	return where;
}
//library substitution
//mutexes
int	netams_lock_init(char *file, unsigned line, u_char type, void *lock, void *attr) {
	mutex_unit *m;

	pthread_mutex_lock(&mutex_lock);
	m = getMutexUnit(lock, ADD);
	m->created = pthread_self();
	pthread_mutex_unlock(&mutex_lock);

	aDebug(DEBUG_MUTEX, "%s:%u %s %p initialized by thread %p\n", \
		file, line, getLockName(type), lock, m->created);

	if(type&LOCK_MUTEX) {
		m->type=LOCK_MUTEX;
		return pthread_mutex_init((pthread_mutex_t*)lock, (const pthread_mutexattr_t*)attr);
	} 
	else if(type&LOCK_RWLOCK) {
		m->type=LOCK_RWLOCK;
		return pthread_rwlock_init((pthread_rwlock_t*)lock, (const pthread_rwlockattr_t*)attr);
	}
	return -1;
}

int	netams_lock_destroy(char *file, unsigned line, u_char type, void *lock) {
	
	pthread_mutex_lock(&mutex_lock);
	getMutexUnit(lock, REMOVE);
	pthread_mutex_unlock(&mutex_lock);
	
	aDebug(DEBUG_MUTEX, "%s %p destroyed by thread %p from %s:%u\n", \
		getLockName(type), lock, pthread_self(), file, line);

	if(type&LOCK_MUTEX) {
		return pthread_mutex_destroy((pthread_mutex_t*)lock);
	} else if(type&LOCK_RWLOCK) {
		return pthread_rwlock_destroy((pthread_rwlock_t*)lock);
	}
	return -1;
}

int	netams_lock_lock(char *file, unsigned line, u_char type, void *lock) {
	mutex_unit *m;
	pthread_t t, caller;

	caller = pthread_self();
	
	pthread_mutex_lock(&mutex_lock);
	m = getMutexUnit(lock, ADD);
	t = getLock(m,ADD);
	pthread_mutex_unlock(&mutex_lock);

	if(t == caller) {
		aDebug(DEBUG_MUTEX, "Thread %p perform %s on %s %p from %s:%u\n", \
			caller, getLockType(type), getLockName(type), lock, file, line);
	} else {
		aLog(D_WARN, "Thread %p tries to %s %s %p locked by %p from %s:%u\n", \
			pthread_self(), getLockType(type), getLockName(type), lock, t, file, line);
		return -1;
	}
	if(type&LOCK_MUTEX)
		return pthread_mutex_lock((pthread_mutex_t*)lock);
	else if(type&LOCK_RWLOCK_RDLOCK) 
		return pthread_rwlock_rdlock((pthread_rwlock_t*)lock);
	else if(type&LOCK_RWLOCK_WRLOCK)
		return pthread_rwlock_wrlock((pthread_rwlock_t*)lock);

	return -1;
}

int	netams_lock_trylock(char *file, unsigned line, u_char type, void *lock) {
	mutex_unit *m;
	pthread_t t, caller;

	caller = pthread_self();
	
	pthread_mutex_lock(&mutex_lock);
	m = getMutexUnit(lock, ADD);
	t = getLock(m,ADD);
	pthread_mutex_unlock(&mutex_lock);

	if(t == caller) {
		aDebug(DEBUG_MUTEX, "Thread %p perform %s on %s  %p from %s:%u\n", \
			caller, getLockType(type), getLockName(type), lock, file, line);
	} else {
		aDebug(DEBUG_MUTEX, "Thread %p %s %p busy for try%s by thread %p from %s:%u\n", \
			caller, getLockName(type), lock, getLockType(type), t, file, line);
	}
	
	if(type&LOCK_MUTEX)
		return pthread_mutex_trylock((pthread_mutex_t*)lock);
	else if(type&LOCK_RWLOCK_RDLOCK)
		return pthread_rwlock_tryrdlock((pthread_rwlock_t*)lock);
	else if(type&LOCK_RWLOCK_WRLOCK)
		return pthread_rwlock_trywrlock((pthread_rwlock_t*)lock);

	return -1;
}

int	netams_lock_unlock(char *file, unsigned line, u_char type, void *lock) {
	mutex_unit *m;
	pthread_t t, caller;

	caller = pthread_self();
	
	pthread_mutex_lock(&mutex_lock);
	m = getMutexUnit(lock, ADD);
	t = getLock(m,REMOVE);
	pthread_mutex_unlock(&mutex_lock);

	if(t == caller) {
		aDebug(DEBUG_MUTEX, "Thread %p perform unlock on %s %p from %s:%u\n", \
			caller, getLockName(type), lock, file, line);
	} else {
		aLog(D_WARN, "Thread %p tries to unlock %s %p not locked by anything from %s:%u\n", \
			pthread_self(), getLockName(type), lock, file, line);
		return -1;
	}

	if(type&LOCK_MUTEX)
		return pthread_mutex_unlock((pthread_mutex_t*)lock);
	else if(type&LOCK_RWLOCK_RDLOCK)
		return pthread_rwlock_unlock((pthread_rwlock_t*)lock);
	else if(type&LOCK_RWLOCK_WRLOCK)
		return pthread_rwlock_unlock((pthread_rwlock_t*)lock);
			        
	return -1;
}

#endif
/////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////
